import base64
from typing import Callable, Dict, Optional

from treelib import Node, Tree

from pytoncenter.address import Address
from pytoncenter.v3.models import Transaction, TransactionTrace

__all__ = [
    "get_opcode",
    "encode_base64",
    "decode_base64",
    "AddressMapping",
    "format_tx",
    "format_trace",
    "truncate_address",
    "create_address_mapping",
]


def get_opcode(data_uint32: int) -> str:
    """
    Get opcode from uint32, the output is a string with 0x prefix, 10 characters long.
    """
    return "0x{:08x}".format(data_uint32).lower()


def encode_base64(hex_data: str) -> str:
    # Convert the hex data back into bytes
    bytes_data = bytes.fromhex(hex_data)
    # Encode these bytes into a base64 string
    base64_encoded = base64.b64encode(bytes_data)
    # Return the base64-encoded string
    return base64_encoded.decode()


def decode_base64(data: str) -> str:
    return base64.b64decode(data + "==").hex()


AddressMapping = Callable[[Address], str]


def _default_address_mapping(address: Address, **kwargs) -> str:
    """
    default_address_mapping map address to a user friendly format with middle part truncated

    Parameters
    ----------
    address : str
        Address to be formatted
    prefix : int, optional
        Prefix length
    suffix : int, optional
        Suffix length
    """
    assert isinstance(address, Address), "address should be an instance of Address"
    prefix = kwargs.get("prefix", 6)
    suffix = kwargs.get("suffix", 6)
    addr = address.to_string(True, is_test_only=False)
    return addr[:prefix] + "..." + addr[-suffix:] if len(addr) > prefix + suffix else addr


def create_address_mapping(mapping: Dict[Address, str], truncate: bool = True) -> AddressMapping:
    def _mapping_func(address: Address) -> str:
        default = truncate_address(address) if truncate else address.to_string(True, is_test_only=False)
        return mapping.get(address, default)

    return _mapping_func


def truncate_address(address: Address, prefix: int = 6, suffix: int = 6) -> str:
    """
    Truncate middle part of the address to make it more readable
    """
    addr = address.to_string(True, is_test_only=False)
    return addr[:prefix] + "..." + addr[-suffix:] if len(addr) > prefix + suffix else addr


def format_tx(tx: Transaction, address_mapping: AddressMapping = _default_address_mapping) -> str:
    tmpl = "\033[95m{src}\033[0m ➡️ \033[92m{dest}\033[0m \033[93m({msg})\033[0m \033[94m[💎 {value} TON]\033[0m"
    src = address_mapping(Address(tx.in_msg.source)) if tx.in_msg.source else "External"
    dst = address_mapping(Address(tx.in_msg.destination)) if tx.in_msg.destination else ""
    value = round((tx.in_msg.value or 0) / 1e9, 5)
    msg = tx.in_msg.opcode
    if (msg is None or msg == "0x00000000") and tx.in_msg.message_content is not None and tx.in_msg.message_content.decoded is not None:
        if tx.in_msg.message_content.decoded.type == "text_comment":
            msg = tx.in_msg.message_content.decoded.comment
        else:
            msg = "<Binary message>"

    return tmpl.format(src=src, dest=dst, msg=msg, value=value)


def format_trace(root: TransactionTrace, address_mapping: AddressMapping = _default_address_mapping) -> str:
    """
    print transaction trace in a pretty way

    Returns
    -------
    str
        The pretty printed transaction trace in tree format
    """

    def recursive_add_node(tree: Tree, trace: TransactionTrace, parent: Optional[Node] = None):
        """
        Recursively add node to the tree
        """
        node = tree.create_node(
            identifier=trace.transaction.hash,
            tag=format_tx(trace.transaction, address_mapping),
            parent=parent,
        )
        for child in trace.children:
            recursive_add_node(tree, child, node)

    tree = Tree()
    recursive_add_node(tree, root)
    return tree.show(stdout=False)  # type: ignore



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar

from tonpy import CellSlice

from pytoncenter.address import Address
from pytoncenter.utils import get_opcode

T = TypeVar("T", bound="BaseMessage")


class BaseMessage(ABC, Generic[T]):
    OPCODE = ""

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @classmethod
    def _preparse(cls, cs: CellSlice) -> CellSlice:
        opcode = get_opcode(cs.load_uint(32))
        assert opcode == cls.OPCODE, f"opcode {opcode} is not {cls.OPCODE}"
        return cs

    @classmethod
    @abstractmethod
    def _parse(cls, body: CellSlice) -> T:
        """
        the cellslice with opcode and body
        """
        raise NotImplementedError

    @classmethod
    def parse(cls, cs: CellSlice) -> T:
        body = cls._preparse(cs)
        return cls._parse(body)

    def to_boc(self) -> str:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{self.name}({self.OPCODE})"


class JettonMessage:
    class InternalTransfer(BaseMessage["InternalTransfer"]):
        OPCODE = "0x178d4519"

        def __init__(
            self,
            query_id: int,
            amount: int,
            sender: Address,
            response_address: Address,
            forward_ton_amount: int,
            forward_payload: Optional[CellSlice],
        ):
            self.query_id = query_id
            self.amount = amount
            self.sender = sender
            self.response_address = response_address
            self.forward_ton_amount = forward_ton_amount
            self.forward_payload = forward_payload

        @classmethod
        def _parse(cls, body: CellSlice):
            """
            internal_transfer  query_id:uint64 amount:(VarUInteger 16) from:MsgAddress
                        response_address:MsgAddress
                        forward_ton_amount:(VarUInteger 16)
                        forward_payload:(Either Cell ^Cell)
            """
            query_id = body.load_uint(64)
            amount = body.load_var_int(16)
            sender = Address(body.load_address())
            response_address = Address(body.load_address())
            forward_ton_amount = body.load_var_uint(16)
            forward_payload = None if body.empty_ext() else body.load_ref(as_cs=True)
            return cls(
                query_id=query_id,
                amount=amount,
                sender=sender,
                response_address=response_address,
                forward_ton_amount=forward_ton_amount,
                forward_payload=forward_payload,
            )

    class Transfer(BaseMessage["Transfer"]):
        OPCODE = "0x0f8a7ea5"

        def __init__(
            self,
            query_id: int,
            amount: int,
            destination: Address,
            response_destination: Address,
            custom_payload: Optional[CellSlice],
            forward_ton_amount: int,
            forward_payload: Optional[CellSlice],
        ):
            self.query_id = query_id
            self.amount = amount
            self.destination = destination
            self.response_destination = response_destination
            self.custom_payload = custom_payload
            self.forward_ton_amount = forward_ton_amount
            self.forward_payload = forward_payload

        @classmethod
        def _parse(cls, body: CellSlice):
            """
            transfer query_id:uint64 amount:(VarUInteger 16) destination:MsgAddress
            response_destination:MsgAddress custom_payload:(Maybe ^Cell)
            forward_ton_amount:(VarUInteger 16) forward_payload:(Either Cell ^Cell)
            """
            query_id = body.load_uint(64)
            amount = body.load_var_uint(16)
            destination = Address(body.load_address())
            response_destination = Address(body.load_address())

            custom_payload = None
            custom_payload_exists = body.load_bool()
            if custom_payload_exists:
                custom_payload = body.load_ref(as_cs=True)

            forward_ton_amount = body.load_var_uint(16)

            forward_payload = None
            if not body.empty_ext():
                forward_payload = body.load_ref(as_cs=True)

            return cls(query_id, amount, destination, response_destination, custom_payload, forward_ton_amount, forward_payload)

    class Excess(BaseMessage["Excess"]):
        OPCODE = "0xd53276db"

        def __init__(self, query_id: int):
            self.query_id = query_id

        @staticmethod
        def _parse(cs: CellSlice) -> JettonMessage.Excess:
            """
            excess query_id:uint64 amount:(VarUInteger 16)
            """
            query_id = cs.load_uint(64)
            return JettonMessage.Excess(query_id)

    class TransferNotification(BaseMessage["TransferNotification"]):
        OPCODE = "0x7362d09c"

        def __init__(self, query_id: int, amount: int, sender: Address, forward_payload: Optional[CellSlice]):
            self.query_id = query_id
            self.amount = amount
            self.sender = sender
            self.forward_payload = forward_payload

        @classmethod
        def _parse(cls, body: CellSlice) -> JettonMessage.TransferNotification:
            """
            transfer_notification query_id:uint64 amount:(VarUInteger 16)
            sender:MsgAddress forward_payload:(Either Cell ^Cell)
            """
            query_id = body.load_uint(64)
            amount = body.load_var_uint(16)
            sender = Address(body.load_address())
            forward_payload = None
            if not body.empty_ext():
                forward_payload = body.load_ref(as_cs=True)
            return cls(query_id, amount, sender, forward_payload)

    class Burn(BaseMessage["Burn"]):
        OPCODE = "0x595f07bc"

        def __init__(
            self,
            query_id: int,
            amount: int,
            response_destination: Address,
            custom_payload: Optional[CellSlice],
        ):
            self.query_id = query_id
            self.amount = amount
            self.response_destination = response_destination
            self.custom_payload = custom_payload

        @classmethod
        def _parse(cls, body: CellSlice) -> JettonMessage.Burn:
            """
            burn query_id:uint64 amount:(VarUInteger 16)
                response_destination:MsgAddress custom_payload:(Maybe ^Cell)
            """
            query_id = body.load_uint(64)
            amount = body.load_var_uint(16)
            response_destination = Address(body.load_address())
            custom_payload = None
            custom_payload_exists = body.load_bool()
            if custom_payload_exists:
                custom_payload = body.load_ref(as_cs=True)
            return cls(query_id, amount, response_destination, custom_payload)

    class BurnNotification(BaseMessage["BurnNotification"]):
        OPCODE = "0x7bdd97de"

        def __init__(self, query_id: int, amount: int, sender: Address, response_destination: Address):
            self.query_id = query_id
            self.amount = amount
            self.sender = sender
            self.response_destination = response_destination

        @classmethod
        def _parse(cls, body: CellSlice) -> JettonMessage.BurnNotification:
            """
            burn_notification query_id:uint64 amount:(VarUInteger 16)
                sender:MsgAddress response_destination:MsgAddress
            """
            query_id = body.load_uint(64)
            amount = body.load_var_uint(16)
            sender = Address(body.load_address())
            response_destination = Address(body.load_address())
            return cls(query_id, amount, sender, response_destination)


class NFTMessage:
    pass



from __future__ import annotations

from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field, field_validator, model_serializer

from .types import AddressLike


class AccountBalance(BaseModel):
    account: AddressLike = Field(..., title="Account")
    balance: str = Field(..., title="Balance")


class BinaryComment(BaseModel):
    type: Literal["binary_comment"] = Field(default="binary_comment", title="Type")
    hex_comment: str = Field(..., title="Hex Comment")


class BlockReference(BaseModel):
    workchain: int = Field(..., title="Workchain")
    shard: str = Field(..., title="Shard")
    seqno: int = Field(..., title="Seqno")


class EstimateFeeRequest(BaseModel):
    address: AddressLike = Field(..., title="Address")
    body: str = Field(..., title="Body")
    init_code: Optional[str] = Field(default=None, title="Init Code")
    init_data: Optional[str] = Field(default=None, title="Init Data")
    ignore_chksig: Optional[bool] = Field(default=True, title="Ignore Chksig")


class ExternalMessage(BaseModel):
    boc: str = Field(
        ...,
        examples=["te6ccgECBQEAARUAAkWIAWTtae+KgtbrX26Bep8JSq8lFLfGOoyGR/xwdjfvpvEaHg"],
        title="Boc",
    )


class Fee(BaseModel):
    in_fwd_fee: int = Field(..., title="In Fwd Fee")
    storage_fee: int = Field(..., title="Storage Fee")
    gas_fee: int = Field(..., title="Gas Fee")
    fwd_fee: int = Field(..., title="Fwd Fee")


GetMethodParameterType = Literal["cell", "slice", "num", "list", "tuple", "unsupported_type"]


class JettonBurn(BaseModel):
    query_id: str = Field(..., title="Query Id")
    owner: AddressLike = Field(..., title="Owner")
    jetton_master: AddressLike = Field(..., title="Jetton Master")
    transaction_hash: str = Field(..., title="Transaction Hash")
    transaction_lt: str = Field(..., title="Transaction Lt")
    transaction_now: int = Field(..., title="Transaction Now")
    response_destination: Optional[str] = Field(..., title="Response Destination")
    custom_payload: Optional[str] = Field(..., title="Custom Payload")


class JettonBurnList(BaseModel):
    jetton_burns: List[JettonBurn] = Field(..., title="Jetton Burns")


class AddressBookEntry(BaseModel):
    user_friendly: str = Field(..., title="User Friendly")


class AddressBook(BaseModel):
    data: Dict[str, AddressBookEntry] = Field(default={}, title="Data")


class JettonContent(BaseModel):
    uri: Optional[str] = Field(None, description="A URI pointing to JSON document with metadata. Used by 'Semi-chain content layout'. ASCII string.")
    name: Optional[str] = Field(None, description="The name of the token - e.g. 'Example Coin'. UTF8 string.")
    description: Optional[str] = Field(None, description="Describes the token - e.g. 'This is an example jetton for the TON network'. UTF8 string.")
    image: Optional[str] = Field(None, description="A URI pointing to a jetton icon with mime type image. ASCII string.")
    image_data: Optional[str] = Field(None, description="Either binary representation of the image for onchain layout or base64 for offchain layout.")
    symbol: Optional[str] = Field(None, description="The symbol of the token - e.g. 'XMPL'. Used in the form 'You received 99 XMPL'. UTF8 string.")
    decimals: Optional[int] = Field(
        9,
        description="The number of decimals the token uses - e.g. 8, means to divide the token amount by 100000000 to get its user representation, while 0 means that tokens are indivisible. If not specified, 9 is used by default. UTF8 encoded string with number from 0 to 255.",
        ge=0,
        le=255,
    )
    amount_style: Optional[Literal["n", "n-of-total", "%"]] = Field(
        "n",
        description="Needed by external applications to understand which format for displaying the number of jettons. 'n' - number of jettons (default value). 'n-of-total' - the number of jettons out of the total number of issued jettons. '%' - percentage of jettons from the total number of issued jettons.",
    )
    render_type: Optional[Literal["currency", "game"]] = Field(
        "currency",
        description="Needed by external applications to understand which group the jetton belongs to and how to display it. 'currency' - display as currency (default value). 'game' - display for games as NFT, considering the amount_style.",
    )


class JettonMaster(BaseModel):
    address: AddressLike = Field(..., title="Address")
    total_supply: int = Field(..., title="Total Supply")
    mintable: bool = Field(..., title="Mintable")
    admin_address: Optional[AddressLike] = Field(..., title="Admin Address")
    last_transaction_lt: int = Field(..., title="Last Transaction Lt")
    jetton_wallet_code_hash: str = Field(..., title="Jetton Wallet Code Hash")
    jetton_content: Optional[JettonContent] = Field(None, title="Jetton Content")
    code_hash: str = Field(..., title="Code Hash")
    data_hash: str = Field(..., title="Data Hash")


class JettonMasterList(BaseModel):
    jetton_masters: List[JettonMaster] = Field(..., title="Jetton Masters")


class JettonTransfer(BaseModel):
    query_id: int = Field(..., title="Query Id")
    source: AddressLike = Field(..., title="Source")
    destination: AddressLike = Field(..., title="Destination")
    amount: int = Field(..., title="Amount")
    source_wallet: AddressLike = Field(..., title="Source Wallet")
    jetton_master: AddressLike = Field(..., title="Jetton Master")
    transaction_hash: str = Field(..., title="Transaction Hash")
    transaction_lt: int = Field(..., title="Transaction Lt")
    transaction_now: int = Field(..., title="Transaction Now")
    response_destination: Optional[AddressLike] = Field(..., title="Response Destination")
    custom_payload: Optional[str] = Field(..., title="Custom Payload")
    forward_ton_amount: Optional[int] = Field(..., title="Forward Ton Amount")
    forward_payload: Optional[str] = Field(..., title="Forward Payload")


class JettonTransferList(BaseModel):
    jetton_transfers: List[JettonTransfer] = Field(..., title="Jetton Transfers")


class JettonWallet(BaseModel):
    address: AddressLike = Field(..., title="Address")
    balance: int = Field(..., title="Balance")
    owner: AddressLike = Field(..., title="Owner")
    jetton: AddressLike = Field(..., title="Jetton")
    last_transaction_lt: int = Field(..., title="Last Transaction Lt")
    code_hash: str = Field(..., title="Code Hash")
    data_hash: str = Field(..., title="Data Hash")


class JettonWalletList(BaseModel):
    jetton_wallets: List[JettonWallet] = Field(..., title="Jetton Wallets")


class MessageInitState(BaseModel):
    hash: str = Field(..., title="Hash")
    body: str = Field(..., title="Body")


class NFTMetadata(BaseModel):
    uri: Optional[str] = Field(default=None, description="Used by `Semi-chain content layout`. ASCII string. A URI pointing to JSON document with metadata.")
    name: Optional[str] = Field(default=None, description="UTF8 string. Identifies the asset.")
    description: Optional[str] = Field(default=None, description="UTF8 string. Describes the asset.")
    image: Optional[str] = Field(default=None, description="ASCII string. A URI pointing to a image with mime type image.")
    image_data: Optional[str] = Field(default=None, description="Either binary representation of the image for onchain layout or base64 for offchain layout.")


class NFTCollection(BaseModel):
    address: AddressLike = Field(..., title="Address")
    owner_address: Optional[AddressLike] = Field(..., title="Owner Address")
    last_transaction_lt: int = Field(..., title="Last Transaction Lt")
    next_item_index: int = Field(..., title="Next Item Index")
    collection_content: Optional[NFTMetadata] = Field(default=None, title="Collection Content")
    code_hash: str = Field(..., title="Code Hash")
    data_hash: str = Field(..., title="Data Hash")


class NFTCollectionList(BaseModel):
    nft_collections: List[NFTCollection] = Field(..., title="Nft Collections")


class NFTItem(BaseModel):
    address: AddressLike = Field(..., title="Address")
    collection_address: Optional[AddressLike] = Field(..., title="Collection Address")
    owner_address: Optional[AddressLike] = Field(..., title="Owner Address")
    init: bool = Field(..., title="Init")
    index: int = Field(..., title="Index")
    last_transaction_lt: int = Field(..., title="Last Transaction Lt")
    code_hash: str = Field(..., title="Code Hash")
    data_hash: str = Field(..., title="Data Hash")
    content: Optional[NFTMetadata] = Field(default=None, title="Content")
    collection: Optional[NFTCollection]


class NFTItemList(BaseModel):
    nft_items: List[NFTItem] = Field(..., title="Nft Items")


class NFTTransfer(BaseModel):
    query_id: int = Field(..., title="Query Id")
    nft_address: AddressLike = Field(..., title="Nft Address")
    transaction_hash: str = Field(..., title="Transaction Hash")
    transaction_lt: int = Field(..., title="Transaction Lt")
    transaction_now: int = Field(..., title="Transaction Now")
    old_owner: AddressLike = Field(..., title="Old Owner")
    new_owner: AddressLike = Field(..., title="New Owner")
    response_destination: Optional[AddressLike] = Field(..., title="Response Destination")
    custom_payload: Optional[str] = Field(..., title="Custom Payload")
    forward_amount: int = Field(..., title="Forward Amount")
    forward_payload: Optional[str] = Field(..., title="Forward Payload")


class NFTTransferList(BaseModel):
    nft_transfers: List[NFTTransfer] = Field(..., title="Nft Transfers")


class SentMessage(BaseModel):
    message_hash: str = Field(
        ...,
        description="Hash of sent message in hex format",
        examples=["383E348617141E35BC25ED9CD0EDEC2A4EAF6413948BF1FB7F865CEFE8C2CD44"],
        title="Message Hash",
    )


class TextComment(BaseModel):
    type: Literal["text_comment"] = Field(default="text_comment", title="Type")
    comment: str = Field(..., title="Comment")


class ValidationError(BaseModel):
    loc: List[Union[str, int]] = Field(..., title="Location")
    msg: str = Field(..., title="Message")
    type: str = Field(..., title="Error Type")


class WalletInfo(BaseModel):
    balance: int = Field(..., title="Balance")
    wallet_type: Optional[str] = Field(..., title="Wallet Type")
    seqno: Optional[int] = Field(..., title="Seqno")
    wallet_id: Optional[int] = Field(..., title="Wallet Id")
    last_transaction_lt: Optional[int] = Field(..., title="Last Transaction Lt")
    last_transaction_hash: Optional[str] = Field(..., title="Last Transaction Hash")
    status: Literal["uninit", "frozen", "active", "nonexist"] = Field(..., title="Status")


class Account(BaseModel):
    balance: int = Field(..., title="Balance")
    code: Optional[str] = Field(..., title="Code")
    data: Optional[str] = Field(..., title="Data")
    last_transaction_lt: Optional[int] = Field(..., title="Last Transaction Lt")
    last_transaction_hash: Optional[str] = Field(..., title="Last Transaction Hash")
    frozen_hash: Optional[str] = Field(default=None, title="Frozen Hash")
    status: Literal["uninit", "frozen", "active", "nonexist"] = Field(..., title="Status")


class AccountState(BaseModel):
    hash: str = Field(..., title="Hash")
    balance: int = Field(default=0, title="Balance")
    account_status: Literal["uninit", "frozen", "active", "nonexist"] = Field(..., title="Account Status")
    frozen_hash: Optional[str] = Field(default=None, title="Frozen Hash")
    code_hash: Optional[str] = Field(..., title="Code Hash")
    data_hash: Optional[str] = Field(..., title="Data Hash")

    @field_validator("account_status", mode="before")
    def set_account_status(cls, v):
        return v or "nonexist"

    @field_validator("balance", mode="before")
    def set_balance(cls, v):
        return int(v or "0")


class Block(BaseModel):
    workchain: int = Field(..., title="Workchain")
    shard: str = Field(..., title="Shard")
    seqno: int = Field(..., title="Seqno")
    root_hash: str = Field(..., title="Root Hash")
    file_hash: str = Field(..., title="File Hash")
    global_id: int = Field(..., title="Global Id")
    version: int = Field(..., title="Version")
    after_merge: bool = Field(..., title="After Merge")
    before_split: bool = Field(..., title="Before Split")
    after_split: bool = Field(..., title="After Split")
    want_merge: bool = Field(..., title="Want Merge")
    want_split: bool = Field(..., title="Want Split")
    key_block: bool = Field(..., title="Key Block")
    vert_seqno_incr: bool = Field(..., title="Vert Seqno Incr")
    flags: int = Field(..., title="Flags")
    gen_utime: int = Field(..., title="Gen Utime")
    start_lt: int = Field(..., title="Start Lt")
    end_lt: int = Field(..., title="End Lt")
    validator_list_hash_short: int = Field(..., title="Validator List Hash Short")
    gen_catchain_seqno: int = Field(..., title="Gen Catchain Seqno")
    min_ref_mc_seqno: int = Field(..., title="Min Ref Mc Seqno")
    prev_key_block_seqno: int = Field(..., title="Prev Key Block Seqno")
    vert_seqno: int = Field(..., title="Vert Seqno")
    master_ref_seqno: Optional[int] = Field(..., title="Master Ref Seqno")
    rand_seed: str = Field(..., title="Rand Seed")
    created_by: str = Field(..., title="Created By")
    tx_count: Optional[int] = Field(..., title="Tx Count")
    masterchain_block_ref: Optional[BlockReference]
    prev_blocks: List[BlockReference] = Field(..., title="Prev Blocks")


class BlockList(BaseModel):
    blocks: List[Block] = Field(..., title="Blocks")


class EstimateFeeResponse(BaseModel):
    source_fees: Fee
    destination_fees: List[Fee] = Field(..., title="Destination Fees")


class GetMethodParameterInput(BaseModel):
    type: GetMethodParameterType = Field(..., title="Type")
    value: Optional[Union[List[GetMethodParameterInput], str, int, bool, AddressLike]] = Field(..., title="Value")

    @model_serializer
    def serialize_value(self):
        if isinstance(self.value, int) and self.type == "num":
            v = hex(self.value)
            return {"type": self.type, "value": v}
        if isinstance(self.value, bool) and self.type == "num":
            v = "-0x1" if self.value else "0x0"
            return {"type": self.type, "value": v}
        return {"type": self.type, "value": self.value}


class GetMethodParameterOutput(BaseModel):
    type: GetMethodParameterType
    value: Optional[Union[List[GetMethodParameterOutput], str]] = Field(..., title="Value")


class HTTPValidationError(BaseModel):
    detail: Optional[List[ValidationError]] = Field(None, title="Detail")


class MasterchainInfo(BaseModel):
    first: Block
    last: Block


class MessageContent(BaseModel):
    hash: str = Field(..., title="Hash")
    body: str = Field(..., title="Body")
    decoded: Optional[Union[TextComment, BinaryComment]] = Field(..., title="Decoded")


class RunGetMethodRequest(BaseModel):
    address: AddressLike = Field(..., title="Address")
    method: str = Field(..., title="Method")
    stack: List[Union[GetMethodParameterInput, Dict[str, Any]]] = Field([], title="Stack")


class RunGetMethodResponse(BaseModel):
    gas_used: int = Field(..., title="Gas Used")
    exit_code: int = Field(..., title="Exit Code")
    stack: List[GetMethodParameterOutput] = Field(..., title="Stack")


class Message(BaseModel):
    hash: str = Field(..., title="Hash")
    source: Optional[str] = Field(..., title="Source")
    destination: Optional[str] = Field(..., title="Destination")
    value: Optional[int] = Field(..., title="Value")
    fwd_fee: Optional[int] = Field(..., title="Fwd Fee")
    ihr_fee: Optional[int] = Field(..., title="Ihr Fee")
    created_lt: Optional[int] = Field(..., title="Created Lt")
    created_at: Optional[int] = Field(..., title="Created At")
    opcode: Optional[str] = Field(..., title="Opcode")
    ihr_disabled: Optional[bool] = Field(..., title="Ihr Disabled")
    bounce: Optional[bool] = Field(..., title="Bounce")
    bounced: Optional[bool] = Field(..., title="Bounced")
    import_fee: Optional[int] = Field(..., title="Import Fee")
    message_content: Optional[MessageContent]
    init_state: Optional[MessageInitState]


class MessageList(BaseModel):
    messages: List[Message] = Field(..., title="Messages")


class Transaction(BaseModel):
    account: AddressLike = Field(..., title="Account")
    hash: str = Field(..., title="Hash")
    lt: int = Field(..., title="Lt")
    now: int = Field(..., title="Now")
    orig_status: Literal["uninit", "frozen", "active", "nonexist"] = Field(..., title="Orig Status")
    end_status: Literal["uninit", "frozen", "active", "nonexist"] = Field(..., title="End Status")
    total_fees: int = Field(..., title="Total Fees")
    prev_trans_hash: str = Field(..., title="Prev Trans Hash")
    prev_trans_lt: int = Field(..., title="Prev Trans Lt")
    description: Any = Field(..., title="Description")
    block_ref: Optional[BlockReference]
    in_msg: Message = Field(..., title="In Msg")
    out_msgs: List[Message] = Field(..., title="Out Msgs")
    account_state_before: Optional[AccountState]
    account_state_after: Optional[AccountState]
    mc_block_seqno: Optional[int] = Field(..., title="Mc Block Seqno")


class TransactionList(BaseModel):
    transactions: List[Transaction] = Field(default=[], title="Transactions")
    address_book: Dict[str, AddressBookEntry] = Field(default={}, title="Address Book")


class TransactionTrace(BaseModel):
    id: str = Field(..., title="Id")
    transaction: Transaction
    children: List[TransactionTrace] = Field(default=[], title="Children")


GetMethodParameterInput.model_rebuild()
GetMethodParameterOutput.model_rebuild()
TransactionTrace.model_rebuild()



from typing import Literal, Optional, overload

from pytoncenter.v2.api import AsyncTonCenterClientV2
from pytoncenter.v3.api import AsyncTonCenterClientV3


@overload
def get_client(version: Literal["v2"], network: Literal["mainnet", "testnet"], qps: Optional[float] = None, *args, **kwargs) -> AsyncTonCenterClientV2: ...


@overload
def get_client(version: Literal["v3"], network: Literal["mainnet", "testnet"], qps: Optional[float] = None, *args, **kwargs) -> AsyncTonCenterClientV3:
    """
    Parameters
    ----------
    network : Union[Literal["mainnet"], Literal["testnet"]]
        The network to use. Only mainnet and testnet are supported.

    api_key : Optional[str], optional
        The API key to use, by default None. If api_key is an empty string, then it will override the environment variable `TONCENTER_API_KEY`.
    custom_endpoint : Optional[str], optional
        The custom endpoint to use. If provided, it will override the network parameter.
    qps: Optional[float], optional
        The maximum queries per second to use. If not provided, it will use 9.5 if api_key is provided, otherwise 1.
    """


def get_client(version: Literal["v2", "v3"], network: Literal["mainnet", "testnet"], *args, **kwargs):
    if version == "v2":
        return AsyncTonCenterClientV2(network=network, *args, **kwargs)
    if version == "v3":
        return AsyncTonCenterClientV3(network=network, *args, **kwargs)
    raise ValueError(f"Invalid version {version}")


__version__ = "0.0.14"



from typing import Callable, Dict, Optional, Union

from deprecated import deprecated
from tonpy import CellSlice
from treelib import Node, Tree

from pytoncenter.address import Address
from pytoncenter.utils import get_opcode

from .types import TraceTx

__all__ = [
    "NamedFunction",
    "truncate_middle",
    "create_named_mapping_func",
    "default_named_func",
    "format_trace_tx",
    "pretty_print_trace_tx",
]

NamedFunction = Callable[[Address], Optional[str]]


@deprecated(version="0.1.0", reason="truncate_middle is deprecated, use `pytoncenter.utils.truncate_address` instead")
def truncate_middle(address: Address, prefix: int = 6, suffix: int = 6) -> str:
    """
    Truncate middle part of the address to make it more readable
    """
    addr = address.to_string(True, is_test_only=False)
    return addr[:prefix] + "..." + addr[-suffix:] if len(addr) > prefix + suffix else addr


@deprecated(version="0.1.0", reason="default_named_func is deprecated, use `pytoncenter.utils.default_address_mapping` instead")
def default_named_func(address: Address, **kwargs) -> Optional[str]:
    """
    Truncate middle part of the address to make it more readable

    Parameters
    ----------
    address : str
        Address to be formatted
    prefix : int, optional
        Prefix length
    suffix : int, optional
        Suffix length
    """
    prefix = kwargs.get("prefix", 6)
    suffix = kwargs.get("suffix", 6)
    return truncate_middle(address, prefix, suffix)


@deprecated(version="0.1.0", reason="format_trace_tx is deprecated, use `pytoncenter.utils.format_tx` instead")
def format_trace_tx(trace_tx: TraceTx, named_func: NamedFunction = default_named_func) -> str:
    """
    Format transaction trace in a pretty way
    A -> B (Value, Opcode)
    """
    raw_src = trace_tx.get("in_msg", {}).get("source", "")
    raw_dst = trace_tx.get("in_msg", {}).get("destination", "")
    value = int(trace_tx.get("in_msg", {}).get("value", "0")) / 1e9
    opcode = "invalid opcode"
    in_msg_data = trace_tx.get("in_msg", {}).get("msg_data", {})
    if in_msg_data.get("@type") == "msg.dataText":
        opcode = trace_tx.get("in_msg", {}).get("message")
    elif in_msg_data.get("@type") == "msg.dataRaw":
        cs = CellSlice(in_msg_data.get("body"))
        opcode = get_opcode(cs.load_uint(32))
    return "{src} -> {dest} ({opcode}) [{value} TON]".format(
        src=named_func(Address(raw_src)) if raw_src else "<external>",
        dest=named_func(Address(raw_dst)),
        value=value,
        opcode=opcode,
    )


@deprecated(version="0.1.0", reason="create_named_mapping_func is deprecated, use `pytoncenter.utils.create_address_mapping` instead")
def create_named_mapping_func(mapping: Dict[Address, str], truncate_address: bool = True) -> NamedFunction:
    """
    create a named function to map address to a more readable format
    For example, you can map address EQAuVtH68PRkM8qsMUh670ecNZPtqs5bANF3MD4dv-eBb2vD to "Alice" using this function
    if the address is not in the mapping, it will return the original address by default
    the original address can be truncated if truncate_address is True
    """

    def namedFunction(address: Union[str, Address]) -> Optional[str]:
        """
        Truncate middle part of the address to make it more readable
        """
        _addr = Address(address)
        default = truncate_middle(_addr, 6, 6) if truncate_address else _addr.to_string(True, is_test_only=False)
        return mapping.get(_addr, default)

    return namedFunction


@deprecated(version="0.1.0", reason="pretty_print_trace_tx with client v2 is deprecated, use `pytoncenter.utils.format_trace` and client v3 instead")
def pretty_print_trace_tx(trace_tx: TraceTx, named_func: NamedFunction = default_named_func):
    """
    print transaction trace in a pretty way
    """

    def recursive_add_tx(tree: Tree, tx: TraceTx, parent: Optional[Node] = None):
        """
        Recursively add node to the tree
        """
        txhash = tx["transaction_id"]["hash"]
        tree.create_node(
            identifier=txhash,
            tag=format_trace_tx(tx, named_func),
            parent=parent,
        )
        for child in tx["children"]:
            recursive_add_tx(tree, child, txhash)

    tree = Tree()
    recursive_add_tx(tree, trace_tx)
    output = tree.show(stdout=False)
    print(output)



class TonException(Exception):
    def __init__(self, code: int):
        self.code = code


class TonCenterException(Exception):
    def __init__(self, code: int, msg: str):
        super().__init__(msg)
        self.code = code
        self.msg = msg

    def __str__(self):
        return f"\033[93mTonCenterException {self.code}: {self.msg}\033[0m"


class TonCenterValidationException(Exception):
    def __init__(self, code: int, msg):
        super().__init__(msg)
        self.code = code
        self.msg = msg

    def __str__(self):
        return f"\033[93mTonCenterValidationException {self.code}: {self.msg}\033[0m"



import asyncio
import os
import uuid
import warnings
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import aiohttp
from tonpy import Cell

from pytoncenter.dispatcher import RoundRobinKeyRotator
from pytoncenter.exception import TonException
from pytoncenter.multicall import Multicallable
from pytoncenter.requestor import AsyncRequestor

from .types import *


class AsyncTonCenterClientV2(Multicallable, AsyncRequestor):
    def __init__(
        self,
        network: Union[Literal["mainnet"], Literal["testnet"]],
        *,
        api_key: Optional[str] = None,
        strategy: Union[Literal["round_robin"], None] = None,
        custom_endpoint: Optional[str] = None,
        qps: Optional[float] = None,
        **kwargs,
    ) -> None:

        self.api_keys = None
        if isinstance(api_key, str):
            if api_key != "":
                self.api_keys = api_key.split(",")
        elif isinstance(api_key, list) and len(api_key) > 0:
            self.api_keys = api_key
        else:
            _api_key = os.getenv("TONCENTER_API_KEY", None)
            if _api_key is not None:
                self.api_keys = _api_key.split(",")

        # show warning if api_key is None
        if not self.api_keys:
            warnings.warn(
                "API key is not provided. TonCenter API is rate limited to 1 request per second. Suggesting providing it in environment variable `TONCENTER_API_KEY` or api_key to increase the rate limit.",
                RuntimeWarning,
            )

        # Key rotation
        self.rotator = None
        if strategy is None:
            if self.api_keys is not None and len(self.api_keys) > 1:
                self.rotator = RoundRobinKeyRotator(keys=self.api_keys)
        elif strategy == "round_robin":
            self.rotator = RoundRobinKeyRotator(keys=self.api_keys)  # type: ignore
        else:
            raise ValueError(f"Strategy {strategy} is not supported")

        assert (network in ["mainnet", "testnet"]) or (custom_endpoint is not None), "Network or custom_endpoint must be provided"
        if custom_endpoint is not None:
            self.base_url = custom_endpoint
        else:
            prefix = "" if network == "mainnet" else "testnet."
            self.base_url = f"https://{prefix}toncenter.com/api/v2"

        # QPS
        if qps is not None:
            assert qps > 0, "QPS must be greater than 0"
        else:
            qps = 9.5 * len(self.api_keys) if self.api_keys else 1
        super().__init__(qps)

    def _get_request_headers(self) -> Dict[str, Any]:
        headers = {
            "Content-Type": "application/json",
            "accept": "application/json",
        }
        _api_key = None
        if self.rotator is not None:
            _api_key = self.rotator.get_key()
        else:
            if self.api_keys is not None:
                _api_key = self.api_keys[0]
        if _api_key:
            headers["X-API-KEY"] = _api_key
        return headers

    async def _parse_response(self, response: aiohttp.ClientResponse):
        """
        _parse_response parses the response from TonCenter API and returns the result if the request was successful.
        If the request was unsuccessful, a TonException is raised.

        Parameters
        ----------
        response : requests.Response
            The response from TonCenter API

        Returns
        -------
        Any
            The result of the request

        Raises
        ------
        aiohttp.client_exceptions.ClientResponseError:
            If the request was unsuccessful, a ClientResponseError is raised.
        TonException
            If the request was successful, but the TonCenter API returned an error, a TonException is raised.
        """
        response.raise_for_status()
        result = await response.json()
        if not result["ok"]:
            raise TonException(result["code"])
        return result["result"]

    async def _async_get(self, handler: str, query: Optional[Dict[str, Any]] = None):
        url = f"{self.base_url}/{handler}"
        params = {k: v for k, v in query.items() if v is not None} if query is not None else None
        return await self._underlying_call("GET", url, params=params)

    async def _async_post(self, handler: str, payload: Optional[Dict[str, Any]] = None):
        url = f"{self.base_url}/{handler}"
        payload = {k: v for k, v in payload.items() if v is not None} if payload is not None else None
        return await self._underlying_call("POST", url, payload=payload)

    async def get_address_information(self, address: str) -> WalletInformation:
        return await self._async_get("getAddressInformation", {"address": address})

    async def get_extended_address_information(self, address: str) -> ExtentedAddressInformation:
        return await self._async_get("getExtendedAddressInformation", {"address": address})

    async def get_wallet_information(self, address: str) -> WalletInformation:
        return await self._async_get("getWalletInformation", {"address": address})

    async def get_address_balance(self, address: str) -> int:
        """
        get_address_balance returns the balance of the address in nanoton.
        """
        balance = await self._async_get("getAddressBalance", {"address": address})
        return int(balance)

    async def get_address_state(self, address: str) -> str:
        """
        get_address_state returns the state of the address as a string, e.g. "active", "uninitialized", etc.
        """
        return await self._async_get("getAddressState", {"address": address})

    async def get_token_data(self, address: str) -> Union[JettonMasterData, JettonWalletData, NFTCollectionData, NFTItemData]:
        return await self._async_get("getTokenData", {"address": address})

    async def pack_address(self, address: str) -> str:
        return await self._async_get("packAddress", {"address": address})

    async def unpack_address(self, address: str) -> str:
        return await self._async_get("unpackAddress", {"address": address})

    async def detect_address(self, address: str) -> DetectAddressResult:
        return await self._async_get("detectAddress", {"address": address})

    async def get_masterchain_info(self) -> MasterChainInfo:
        return await self._async_get("getMasterchainInfo")

    async def get_masterchain_block_signatures(self, seq_no: int) -> BlockSignatures:
        return await self._async_get("getMasterchainBlockSignatures", {"seq_no": seq_no})

    async def get_shards(self, seqno: int) -> Shards:
        return await self._async_get("shards", {"seqno": seqno})

    async def get_shard_block_proof(self, workchain: int, shard: int, seqno: int, from_seqno: Optional[int] = None):
        query = {"workchain": workchain, "shard": shard, "seqno": seqno}
        if from_seqno is not None:
            query["from_seqno"] = from_seqno
        return await self._async_get("getShardBlockProof", query=query)

    async def get_consensus_block(self) -> ConsensusBlock:
        return await self._async_get("getConsensusBlock")

    async def lookup_block(
        self,
        workchain: int,
        shard: int,
        seqno: Optional[int] = None,
        lt: Optional[int] = None,
        unixtime: Optional[int] = None,
    ):
        query = {"workchain": workchain, "shard": shard}
        if seqno is not None:
            query["seqno"] = seqno
        if lt is not None:
            query["lt"] = lt
        if unixtime is not None:
            query["unixtime"] = unixtime
        return await self._async_get("lookupBlock", query=query)

    async def shards(self, seqno: int):
        return await self._async_get("shards", {"seqno": seqno})

    async def get_block_transactions(
        self,
        workchain: int,
        shard: int,
        seqno: int,
        root_hash: Optional[str] = None,
        file_hash: Optional[str] = None,
        after_lt: Optional[int] = None,
        after_hash: Optional[str] = None,
        count: Optional[int] = None,
    ):
        query = {"workchain": workchain, "shard": shard, "seqno": seqno}
        if root_hash is not None:
            query["root_hash"] = root_hash
        if file_hash is not None:
            query["file_hash"] = file_hash
        if after_lt is not None:
            query["after_lt"] = after_lt
        if after_hash is not None:
            query["after_hash"] = after_hash
        if count is not None:
            query["count"] = count
        return await self._async_get("getBlockTransactions", query=query)

    async def get_block_header(
        self,
        workchain: int,
        shard: int,
        seqno: int,
        root_hash: Optional[str] = None,
        file_hash: Optional[str] = None,
    ):
        query = {"workchain": workchain, "shard": shard, "seqno": seqno}
        if root_hash is not None:
            query["root_hash"] = root_hash
        if file_hash is not None:
            query["file_hash"] = file_hash
        return await self._async_get("getBlockHeader", query=query)

    async def get_transactions(
        self,
        address: str,
        latest_lt: Optional[int] = None,
        limit: Optional[int] = None,
        begin_lt: Optional[int] = None,
        hash: Optional[str] = None,
        archival: bool = True,
    ) -> List[Tx]:
        """
        get_transactions returns the transactions of the address in descending order of logical time

        Parameters
        ----------
        address : str
            The address to get transactions
        limit : Optional[int], optional
            The limit of transactions to get, default maximum 100
        latest_lt : int, optional
            The latest logical time to get transactions, default None means always find the latest transactions
        lt : Optional[int], optional
            The logical time of the transaction to start getting transactions
        hash : Optional[str], optional
            The hash of the transaction to start getting transactions
        """
        assert limit is None or limit <= 100, "Limit must be less than or equal to 100"
        assert (begin_lt != 0 and hash != "") or (begin_lt == 0 and hash == ""), "begin_lt and hash must be specified together"
        return await self._async_get(
            "getTransactions",
            {
                "address": address,
                "limit": limit,
                "lt": begin_lt,
                "hash": hash,
                "to_lt": latest_lt,
                "archival": "false" if not archival else "true",
            },
        )

    async def try_locate_tx(self, source: str, destination: str, created_lt: int) -> Tx:
        return await self._async_get(
            "tryLocateTx",
            {"source": source, "destination": destination, "created_lt": created_lt},
        )

    async def try_locate_result_tx(self, source: str, destination: str, created_lt: int) -> Tx:
        return await self._async_get(
            "tryLocateResultTx",
            {"source": source, "destination": destination, "created_lt": created_lt},
        )

    async def try_locate_source_tx(self, source: str, destination: str, created_lt: int) -> Tx:
        return await self._async_get(
            "tryLocateSourceTx",
            {"source": source, "destination": destination, "created_lt": created_lt},
        )

    async def get_config_param(self, config_id: int, seqno: Optional[int] = None):
        return await self._async_get("getConfigParam", {"config_id": config_id, "seqno": seqno})

    async def run_get_method(self, address: str, method_name: str, stack: List[Tuple[str, Any]] = []) -> GetMethodResult:
        result = await self._async_post("runGetMethod", {"address": address, "method": method_name, "stack": stack})
        if result.get("@type") == "smc.runResult" and "stack" in result:
            r: Dict[str, Any] = result["stack"]
            return [{"type": r[i][0], "value": r[i][1]} for i in range(len(r))]
        raise ValueError(f"Invalid get method result: {result}")

    async def send_boc(self, boc: str):
        return await self._async_post("sendBoc", {"boc": boc})

    async def send_boc_return_hash(self, boc: str):
        return await self._async_post("sendBocReturnHash", {"boc": boc})

    async def send_query(self, address: str, body: str, init_code: str, init_data: str):
        return await self._async_post(
            "sendQuery",
            {
                "address": address,
                "body": body,
                "init_code": init_code,
                "init_data": init_data,
            },
        )

    async def estimate_fee(
        self,
        address: str,
        body: Union[str, Cell],
        init_code: str,
        init_data: str,
        ignore_chksig: bool = True,
    ) -> EstimateResult:
        if isinstance(body, Cell):
            body = body.to_boc()
        return await self._async_post(
            "estimateFee",
            {
                "address": address,
                "body": body,
                "init_code": init_code,
                "init_data": init_data,
                "ignore_chksig": int(ignore_chksig),
            },
        )

    async def json_rpc(
        self,
        method: str,
        params: Dict[str, Any],
        id: Optional[int] = None,
        jsonrpc: Optional[str] = "2.0",
    ):
        id = uuid.uuid4().int if id is None else id
        return await self._async_post(
            "jsonRpc",
            {"method": method, "params": params, "id": id, "jsonrpc": jsonrpc},
        )

    async def subscribe_tx(self, address: str, interval_in_second: float = 1.0, limit: int = 20):
        """
        subscribeTx returns a generator that yields transactions of the address

        Parameters
        ----------
        address : str
            The address to subscribe to
        interval_in_second : float
            The interval_in_second to poll the TonCenter API, by default 1 second
        limit : int
            The limit of transactions to get, by default 20. Means the maximum number of transactions to get in one request.
            And the oldest transaction you can get in default is 20 transactions from the latest transaction.

        Yields
        ------
        Tx
            The transaction of the address
        """
        cur_lt = None
        while True:
            try:
                if cur_lt is not None:
                    await asyncio.sleep(interval_in_second)
                results = await self.get_transactions(address, limit=limit, latest_lt=cur_lt)
                if not results:
                    continue

                # The results will be sorted by lt in descending order by default, so we need to reverse it
                results.reverse()

                # Now the last transaction is the latest transaction, it will be the next begin_lt and begin_txhash
                cur_lt = results[-1]["transaction_id"]["lt"]

                for tx in results:
                    yield tx

            except asyncio.CancelledError:
                break
            except KeyboardInterrupt:
                break
            except Exception as e:
                print(e)

    async def trace_tx(self, root_tx: Tx) -> TraceTx:
        """
        trace_tx traces the transaction and its children transactions

        Example
        -------
        ```
        client = AsyncTonCenterClient(network="testnet")
        txs = await client.get_transactions("kQAreQ23eabjRO5glLCbhZ4KxQ9SOIjtw2eM2PuEXXhIZeh3", limit=1)
        trace = await client.trace_tx(txs[0])
        pprint(trace)
        """

        async def get_children_tx(source: str, destination: str, created_lt: int) -> Optional[TraceTx]:
            try:
                _tx = await self.try_locate_tx(source=source, destination=destination, created_lt=created_lt)
                _trace: TraceTx = {
                    "@type": _tx["@type"],
                    "address": _tx["address"],
                    "data": _tx["data"],
                    "fee": _tx["fee"],
                    "other_fee": _tx["other_fee"],
                    "storage_fee": _tx["storage_fee"],
                    "transaction_id": _tx["transaction_id"],
                    "utime": _tx["utime"],
                    "in_msg": _tx["in_msg"],
                    "children": [],
                }
                tasks = [
                    get_children_tx(
                        source=msg["source"],
                        destination=msg["destination"],
                        created_lt=int(msg["created_lt"]),
                    )
                    for msg in _tx["out_msgs"]
                ]
                _trace["children"] = [child_tx for child_tx in await asyncio.gather(*tasks) if child_tx is not None]
                return _trace
            except TonException:
                return None

        output: TraceTx = {
            "@type": root_tx["@type"],
            "address": root_tx["address"],
            "data": root_tx["data"],
            "fee": root_tx["fee"],
            "other_fee": root_tx["other_fee"],
            "storage_fee": root_tx["storage_fee"],
            "transaction_id": root_tx["transaction_id"],
            "utime": root_tx["utime"],
            "in_msg": root_tx["in_msg"],
            "children": [],
        }

        tasks = [
            get_children_tx(
                source=msg["source"],
                destination=msg["destination"],
                created_lt=int(msg["created_lt"]),
            )
            for msg in root_tx["out_msgs"]
        ]
        output["children"] = [child_tx for child_tx in await asyncio.gather(*tasks) if child_tx is not None]

        return output



import asyncio
from typing import Any, Coroutine, Dict, Iterable, List, TypeVar, overload

T = TypeVar("T")
K = TypeVar("K")


class Multicallable:
    @overload
    async def multicall(self, *coros: Coroutine[Any, Any, T]) -> List[T]: ...

    @overload
    async def multicall(self, coros: Dict[K, Coroutine[Any, Any, T]]) -> Dict[K, T]: ...

    @overload
    async def multicall(self, coros: List[Coroutine[Any, Any, T]]) -> List[T]: ...

    async def multicall(self, *args, **kwargs) -> Any:
        """
        With multicall, you can execute multiple async function calls by this method.
        This function will return a list of results or a dictionary of results if you pass a dictionary of coroutines.


        Code Snippet
        ------------
        ```python
        client = get_client("v2 or v3", network="testnet or mainnet")
        result = await client.multicall(
            any_asnyc_func1(param1, param2...),
            any_asnyc_func2(param1, param2...),
            any_asnyc_func3(param1, param2...),
        )
        print(result)
        ```

        Example1
        -------
        ```python
        client = get_client("v2", network="testnet")
        result = await client.multicall([
            client.get_address_information("address1"),
            client.get_address_information("address2"),
        ])
        print(result)
        ```

        Example2
        -------
        ```python
        client = get_client("v2", network="testnet")
        result = await client.multicall(
            client.get_address_balance("address1"),
            client.run_get_method("kQBqSpvo4S87mX9tjHaG4zhYZeORhVhMapBJpnMZ64jhrP-A", "get_jetton_data", {})
        )
        print(result)
        ```

        Example3
        -------
        ```python
        client = get_client("v2", network="testnet")
        result = await client.multicall({
            "task1": client.get_address_balance("address1"),
            "task2": client.run_get_method("kQBqSpvo4S87mX9tjHaG4zhYZeORhVhMapBJpnMZ64jhrP-A", "get_jetton_data", {})
        })
        print(result)
        ```
        """

        if len(args) == 1 and isinstance(args[0], (dict, list)):
            coros = args[0]
        else:
            # check if the arguments are coroutines
            assert all([asyncio.iscoroutine(coro) for coro in args]), "All element should be Coroutines"
            coros = args

        if isinstance(coros, dict):
            tasks = {name: asyncio.create_task(coro) for name, coro in coros.items()}
            results = await asyncio.gather(*tasks.values(), return_exceptions=True)
            return {name: result for name, result in zip(tasks.keys(), results)}
        elif isinstance(coros, Iterable):
            tasks = [asyncio.create_task(coro) for coro in coros]
            return await asyncio.gather(*tasks, return_exceptions=True)
        else:
            raise ValueError("Invalid argument type")



import asyncio
import hashlib
import os
import time
import warnings
from typing import Any, Dict, List, Literal, Optional, Tuple, Union, overload

import aiohttp
from tonpy import CellSlice, begin_cell

from pytoncenter.address import Address
from pytoncenter.dispatcher import RoundRobinKeyRotator
from pytoncenter.exception import TonCenterException, TonCenterValidationException
from pytoncenter.multicall import Multicallable
from pytoncenter.requestor import AsyncRequestor
from pytoncenter.v3.models import *


class AsyncTonCenterClientV3(Multicallable, AsyncRequestor):
    def __init__(
        self,
        network: Union[Literal["mainnet"], Literal["testnet"]],
        *,
        api_key: Union[None, str, List[str]] = None,
        strategy: Union[Literal["round_robin"], None] = None,
        custom_endpoint: Optional[str] = None,
        qps: Optional[float] = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        network : Union[Literal["mainnet"], Literal["testnet"]]
            The network to use. Only mainnet and testnet are supported.

        api_key : Optional[str], optional
            The API key to use, by default None. If api_key is an empty string, then it will override the environment variable `TONCENTER_API_KEY`.
            - If api key is None, it will use the environment variable `TONCENTER_API_KEY`. You can provide in a comma separated string to use multiple keys.
            - If api key is an empty string, it will not use any API key
            - If api key is a string, it will use the provided API key
            - If api key is a list of strings, it will use the round robin strategy to rotate the keys
        strategy : Union[Literal["round_robin"], None], optional
            The strategy to use for rotating the API keys. if len(api_key) > 1. round_robin will be used by default.
        custom_endpoint : Optional[str], optional
            The custom endpoint to use. If provided, it will override the network parameter.
        qps: Optional[float], optional
            The maximum queries per second to use. If not provided, it will use 9.5 * len(api keys) if api_key is provided, otherwise 1.
        """
        self._network = network
        # API KEY
        self.api_keys = None
        if isinstance(api_key, str):
            if api_key != "":
                self.api_keys = api_key.split(",")
        elif isinstance(api_key, list) and len(api_key) > 0:
            self.api_keys = api_key
        else:
            _api_key = os.getenv("TONCENTER_API_KEY", None)
            if _api_key is not None:
                self.api_keys = _api_key.split(",")
        # show warning if api_key is None
        if not self.api_keys:
            warnings.warn(
                "API key is not provided. TonCenter API is rate limited to 1 request per second. Suggesting providing it in environment variable `TONCENTER_API_KEY` or api_key to increase the rate limit.",
                RuntimeWarning,
            )

        # Key rotation
        self.rotator = None
        if strategy is None:
            if self.api_keys is not None and len(self.api_keys) > 1:
                self.rotator = RoundRobinKeyRotator(keys=self.api_keys)
        elif strategy == "round_robin":
            self.rotator = RoundRobinKeyRotator(keys=self.api_keys)  # type: ignore
        else:
            raise ValueError(f"Strategy {strategy} is not supported")

        # Network and custom endpoint
        assert (network in ["mainnet", "testnet"]) or (custom_endpoint is not None), "Network or custom_endpoint must be provided"
        if custom_endpoint is not None:
            self.base_url = custom_endpoint
        else:
            prefix = "" if network == "mainnet" else "testnet."
            self.base_url = f"https://{prefix}toncenter.com/api/v3"

        # QPS
        if qps is not None:
            assert qps > 0, "QPS must be greater than 0"
        else:
            qps = 9.5 * len(self.api_keys) if self.api_keys else 1
        super().__init__(qps)

    def _get_request_headers(self) -> Dict[str, Any]:
        headers = {
            "Content-Type": "application/json",
            "accept": "application/json",
        }
        _api_key = None
        if self.rotator is not None:
            _api_key = self.rotator.get_key()
        else:
            if self.api_keys is not None:
                _api_key = self.api_keys[0]
        if _api_key:
            headers["X-API-KEY"] = _api_key
        return headers

    async def _parse_response(self, response: aiohttp.ClientResponse) -> Dict[str, Any]:
        """
        _parse_response parses the response from TonCenter API and returns the result if the request was successful.
        If the request was unsuccessful, a TonException is raised.

        Parameters
        ----------
        response : requests.Response
            The response from TonCenter API

        Returns
        -------
        Any
            The result of the request

        Raises
        ------
        aiohttp.client_exceptions.ClientResponseError:
            If the request was unsuccessful, a ClientResponseError is raised.
        TonException
            If the request was successful, but the TonCenter API returned an error, a TonException is raised.
        """

        result = await response.json()
        if not response.ok:
            if response.status == 422:
                raise TonCenterValidationException(response.status, HTTPValidationError(detail=result.get("detail")))
            else:
                raise TonCenterException(response.status, result.get("error"))
        return result

    async def _async_get(self, handler: str, query: Optional[Dict[str, Any]] = None):
        url = f"{self.base_url}/{handler}"
        params = {k: v for k, v in query.items() if v is not None} if query is not None else None
        return await self._underlying_call("GET", url, params=params)

    async def _async_post(self, handler: str, payload: Dict[str, Any]):
        url = f"{self.base_url}/{handler}"
        payload = {k: v for k, v in payload.items() if v is not None}
        return await self._underlying_call("POST", url, payload=payload)

    async def get_masterchain_info(self) -> MasterchainInfo:
        resp = await self._async_get("masterchainInfo")
        return MasterchainInfo(**resp)

    async def get_blocks(self, req: Optional[GetBlockRequest] = None) -> List[Block]:
        req = req if req is not None else GetBlockRequest()
        resp = await self._async_get("blocks", req.model_dump(exclude_none=True))
        blk_list = BlockList(**resp)
        return blk_list.blocks

    async def get_masterchain_block_shards(self, req: GetMasterchainBlockShardsRequest) -> List[Block]:
        resp = await self._async_get("masterchainBlockShards", req.model_dump(exclude_none=True))
        blk_list = BlockList(**resp)
        return blk_list.blocks

    async def get_masterchain_block_shard_state(self, req: GetMasterchainBlockShardStateRequest) -> List[Block]:
        resp = await self._async_get("masterchainBlockShardState", req.model_dump(exclude_none=True))
        b = BlockList(**resp)
        return b.blocks

    async def get_address_book(self, req: GetAddressBookRequest) -> Dict[str, AddressBookEntry]:
        """
        get_address_book returns the mapping for original address to user friendly format
        """
        resp = await self._async_get("addressBook", req.model_dump(exclude_none=True))
        addr_book = AddressBook(data=resp)
        return addr_book.data

    @overload
    async def get_transactions(self, req: Optional[GetTransactionByHashRequest]) -> Union[Tuple[None, None], Tuple[Transaction, Dict[str, AddressBookEntry]]]: ...

    @overload
    async def get_transactions(self, req: Optional[GetTransactionsRequest]) -> Tuple[List[Transaction], Dict[str, AddressBookEntry]]: ...

    async def get_transactions(
        self, req: Union[Optional[GetTransactionsRequest], GetTransactionByHashRequest] = None
    ) -> Union[Tuple[None, None], Tuple[Transaction, Dict[str, AddressBookEntry]], Tuple[List[Transaction], Dict[str, AddressBookEntry]]]:
        req = req or GetTransactionsRequest()
        resp = await self._async_get("transactions", req.model_dump(exclude_none=True))
        tx_list = TransactionList(**resp)
        if isinstance(req, GetTransactionByHashRequest):
            if len(tx_list.transactions) == 0:
                return None, None
            assert len(tx_list.transactions) == 1, "The response should contain only one transaction"
            return tx_list.transactions[0], tx_list.address_book
        return tx_list.transactions, tx_list.address_book

    async def get_transactions_by_masterchain_block(self, req: GetTransactionByMasterchainBlockRequest) -> Tuple[List[Transaction], Dict[str, AddressBookEntry]]:
        resp = await self._async_get("transactionsByMasterchainBlock", req.model_dump(exclude_none=True))
        tx_list = TransactionList(**resp)
        return tx_list.transactions, tx_list.address_book

    async def get_transaction_by_message(self, req: GetTransactionByMessageRequest) -> Tuple[List[Transaction], Dict[str, AddressBookEntry]]:
        resp = await self._async_get("transactionsByMessage", req.model_dump(exclude_none=True))
        tx_list = TransactionList(**resp)
        return tx_list.transactions, tx_list.address_book

    async def get_adjacent_transactions(self, req: GetAdjacentTransactionsRequest) -> Tuple[List[Transaction], Dict[str, AddressBookEntry]]:
        try:
            if req.full is False:
                resp = await self._async_get("adjacentTransactions", req.model_dump(exclude_none=True))
                tx_list = TransactionList(**resp)
                return tx_list.transactions, tx_list.address_book
            else:
                address_book = dict()
                transactions = []
                while True:
                    resp = await self._async_get("adjacentTransactions", req.model_dump(exclude_none=True))
                    txs = TransactionList(**resp)
                    transactions.extend(txs.transactions)
                    address_book.update(txs.address_book)
                    if len(txs.transactions) < req.limit:
                        break
                    req.offset += req.limit
                return transactions, address_book
        except TonCenterException as e:
            if e.code == 404:
                return [], {}
            raise e

    @overload
    async def get_messages(self, req: GetMessageByHashRequest) -> Optional[Message]: ...

    @overload
    async def get_messages(self, req: Optional[GetMessagesRequest] = None) -> List[Message]: ...

    async def get_messages(self, req: Union[Optional[GetMessagesRequest], GetMessageByHashRequest] = None) -> Union[List[Message], Optional[Message]]:
        req = req if req is not None else GetMessagesRequest()
        resp = await self._async_get("messages", req.model_dump(exclude_none=True))
        msg_list = MessageList(**resp)
        if isinstance(req, GetMessageByHashRequest):
            if len(msg_list.messages) == 0:
                return None
            assert len(msg_list.messages) == 1, "The response should contain only one message"
            return msg_list.messages[0]
        return msg_list.messages

    async def get_nft_collections(self, req: Optional[GetNFTCollectionsRequest] = None) -> List[NFTCollection]:
        req = req if req is not None else GetNFTCollectionsRequest()
        resp = await self._async_get("nft/collections", req.model_dump(exclude_none=True))
        collections = NFTCollectionList(**resp)
        return collections.nft_collections

    @overload
    async def get_nft_items(self, req: Optional[GetNFTItemsRequest] = None) -> List[NFTItem]: ...

    @overload
    async def get_nft_items(self, req: GetSpecifiedNFTItemRequest) -> Optional[NFTItem]: ...

    async def get_nft_items(self, req: Optional[Union[GetNFTItemsRequest, GetSpecifiedNFTItemRequest]] = None) -> Union[List[NFTItem], Optional[NFTItem]]:
        req = req if req is not None else GetNFTItemsRequest()
        resp = await self._async_get("nft/items", req.model_dump(exclude_none=True))
        items = NFTItemList(**resp)
        if isinstance(req, GetSpecifiedNFTItemRequest):
            if len(items.nft_items) == 0:
                return None
            assert len(items.nft_items) == 1, "The response should contain only one item"
            return items.nft_items[0]
        return items.nft_items

    async def get_nft_transfers(self, req: Optional[GetNFTTransfersRequest] = None) -> List[NFTTransfer]:
        req = req if req is not None else GetNFTTransfersRequest()
        resp = await self._async_get("nft/transfers", req.model_dump(exclude_none=True))
        transfers = NFTTransferList(**resp)
        return transfers.nft_transfers

    @overload
    async def get_jetton_masters(self, req: Optional[GetJettonMastersRequest] = None) -> List[JettonMaster]:
        """
        get_jetton_masters returns the jetton masters.
        """

    @overload
    async def get_jetton_masters(self, req: AddressLike) -> Optional[JettonMaster]:
        """
        get_jetton_masters returns the specified jetton master with its address as the parameter
        """

    async def get_jetton_masters(self, req: Union[Optional[GetJettonMastersRequest], AddressLike] = None) -> Union[List[JettonMaster], Optional[JettonMaster]]:
        req = req if req is not None else GetJettonMastersRequest()
        is_address = isinstance(req, (str, Address))
        if is_address:
            req = GetJettonMastersRequest(address=req)
        resp = await self._async_get("jetton/masters", req.model_dump(exclude_none=True))
        masters = JettonMasterList(**resp)
        if is_address:
            if len(masters.jetton_masters) == 0:
                return None
            assert len(masters.jetton_masters) == 1, "The response should contain only one master"
            return masters.jetton_masters[0]
        return masters.jetton_masters

    @overload
    async def get_jetton_wallets(self, req: Optional[GetJettonWalletsRequest] = None) -> List[JettonWallet]: ...

    @overload
    async def get_jetton_wallets(self, req: GetSpecifiedJettonWalletRequest) -> Optional[JettonWallet]: ...

    async def get_jetton_wallets(self, req: Optional[Union[GetJettonWalletsRequest, GetSpecifiedJettonWalletRequest]] = None) -> Union[List[JettonWallet], Optional[JettonWallet]]:
        req = req if req is not None else GetJettonWalletsRequest()
        resp = await self._async_get("jetton/wallets", req.model_dump(exclude_none=True))
        wallets = JettonWalletList(**resp)
        if isinstance(req, GetSpecifiedJettonWalletRequest):
            if len(wallets.jetton_wallets) == 0:
                return None
            assert len(wallets.jetton_wallets) == 1, "The response should contain only one wallet"
            return wallets.jetton_wallets[0]
        return wallets.jetton_wallets

    async def get_jetton_transfers(self, req: Optional[GetJettonTransfersRequest] = None) -> List[JettonTransfer]:
        req = req if req is not None else GetJettonTransfersRequest()
        resp = await self._async_get("jetton/transfers", req.model_dump(exclude_none=True))
        txfer = JettonTransferList(**resp)
        return txfer.jetton_transfers

    async def get_jetton_burns(self, req: Optional[GetJettonBurnsRequest] = None) -> List[JettonBurn]:
        req = req if req is not None else GetJettonBurnsRequest()
        resp = await self._async_get("jetton/burns", req.model_dump(exclude_none=True))
        burns = JettonBurnList(**resp)
        return burns.jetton_burns

    async def get_account(self, req: GetAccountRequest) -> Account:
        resp = await self._async_get("account", req.model_dump(exclude_none=True))
        return Account(**resp)

    async def get_wallet(self, req: GetWalletRequest) -> WalletInfo:
        resp = await self._async_get("wallet", req.model_dump(exclude_none=True))
        return WalletInfo(**resp)

    async def send_message(self, req: ExternalMessage) -> SentMessage:
        resp = await self._async_post("message", req.model_dump(exclude_none=True))
        return SentMessage(**resp)

    async def run_get_method(self, req: RunGetMethodRequest) -> RunGetMethodResponse:
        resp = await self._async_post("runGetMethod", req.model_dump(exclude_none=True))
        return RunGetMethodResponse(**resp)

    async def estimate_fee(self, req: EstimateFeeRequest) -> EstimateFeeResponse:
        resp = await self._async_post("estimateFee", req.model_dump(exclude_none=True))
        return EstimateFeeResponse(**resp)

    async def get_dns_record(self, req: GetDNSRecordRequest) -> DNSRecord:
        """
        get_dns_record returns the DNS record for the given domain name and category.
        Only mainnet is supported.
        Only the wallet category is supported at the moment.
        """
        assert self._network == "mainnet", "Only mainnet is supported"
        dns_name = req.dns_name.replace(".", "\0")
        dns_slice = begin_cell().store_string(dns_name).end_cell().to_boc()
        category_hash = 0
        if req.category is not None:
            category_hash = int(hashlib.sha256(req.category.encode()).hexdigest(), 16)
        resp = await self.run_get_method(
            RunGetMethodRequest(
                address="EQC3dNlesgVD8YbAazcauIrXBPfiVhMMr5YYk2in0Mtsz0Bz",
                method="dnsresolve",
                stack=[
                    GetMethodParameterInput(type="slice", value=dns_slice),
                    GetMethodParameterInput(type="num", value=category_hash),
                ],
            )
        )
        # We cannot import decoder here because it will cause circular import
        # Thus we need to handle the decoding here
        # It might be dirty, will find a better way to handle this
        assert len(resp.stack) == 2, "Expecting to find two items in the stack"
        assert resp.stack[0].type == "num", "Expecting the first item to be a number"
        assert resp.stack[1].type == "cell", "Expecting the second item to be a cell"
        subdomain_bits = int(resp.stack[0].value, 16)  # type: ignore
        cs = CellSlice(resp.stack[1].value)  # type: ignore
        if req.category == "wallet":
            _ = cs.load_uint(16)  # opcode
            try:
                address = cs.load_address()
            except:
                address = None
            return DNSRecord(
                subdomain_bits=subdomain_bits,
                address=address,
            )
        raise NotImplementedError(f"Decoding {req.category} category is not implemented yet")

    async def wait_message_exists(self, req: WaitMessageExistsRequest):
        """
        wait_message_exists wait until the whole transaction trace is complete and yields the transaction.
        This is useful after the external message is sent, and we want to use the message hash to get the transaction trace.
        """
        retry = req.max_retry
        while retry is None or retry > 0:
            retry = retry - 1 if retry is not None else None
            _timer_start = time.monotonic()
            try:
                msgs, _ = await self.get_transaction_by_message(
                    GetTransactionByMessageRequest(
                        direction="in",
                        msg_hash=req.msg_hash,
                        limit=1,
                    )
                )
            except TonCenterException as e:
                if e.code == 503:
                    msgs = []
                else:
                    raise e
            _timer_end = time.monotonic()
            elapse = _timer_end - _timer_start
            sleep_time = max(0, req.interval - elapse)
            if len(msgs) == 0:
                await asyncio.sleep(sleep_time)
                continue
            assert len(msgs) == 1, f"Expecting to find one transaction by message hash {req.msg_hash}, but found {len(msgs)}"
            yield msgs[0]
            return
        raise TonCenterException(429, "Reached the maximum retry limit")

    async def subscribe_tx(self, req: SubscribeTransactionRequest):
        """
        subscribe_tx subscribes to transactions of a wallet and yields the transactions as they come.
        """
        warnings.warn("\033[93mThe `subscribe_tx` function is currently under development; please use it with caution.\033[0m", UserWarning)

        while True:
            _timer_start = time.monotonic()
            txs, _ = await self.get_transactions(
                GetTransactionsRequest(
                    account=req.account,
                    start_utime=req.start_time,
                    sort="asc",
                    limit=req.limit,
                    offset=req.offset,
                )
            )
            req.offset += len(txs)
            if len(txs) > 0:
                for tx in txs:
                    yield tx
            _timer_end = time.monotonic()
            elapse = _timer_end - _timer_start
            sleep_time = max(0, req.interval - elapse)
            await asyncio.sleep(sleep_time)

    async def get_trace_alternative(self, req: GetTransactionTraceRequest) -> TransactionTrace:
        """
        get_trace_alternatives takes a transaction hash as input and returns the transaction trace.

        # Note
        This is an alternative method to get the transaction trace. It is not recommended to use this method in production unless the
        original method does not work. It is compatible with the original method, but it may not be as efficient as it.
        """

        # Trace source of the transaction
        async def _trace_source(orig_tx: Transaction) -> Transaction:
            """
            Find the source of the transaction, and return all the transaction that we found.
            External message are always the source of the transaction.

            Returns
            -------
            Transaction
                The source transaction
            """
            current_tx = orig_tx
            while current_tx.in_msg.source is not None:
                candidates, _ = await self.get_adjacent_transactions(
                    GetAdjacentTransactionsRequest(
                        hash=current_tx.hash,
                        direction="in",
                        limit=1,
                    )
                )
                assert len(candidates) == 1, f"Expecting to find one transaction by message hash {current_tx.in_msg.hash}, but found {len(candidates)}"
                prev_tx = candidates[0]
                current_tx = prev_tx
            return current_tx

        async def _dfs_trace(root: Transaction, trace_id: str) -> List[TransactionTrace]:
            next_txs, _ = await self.get_adjacent_transactions(GetAdjacentTransactionsRequest(hash=root.hash, direction="out", limit=256, sort=req.sort, full=True))
            results = await self.multicall([_dfs_trace(tx, trace_id) for tx in next_txs])
            return [TransactionTrace(id=tx.hash, transaction=tx, children=results[i]) for i, tx in enumerate(next_txs)]

        orig_tx, _ = await self.get_transactions(GetTransactionByHashRequest(hash=req.hash))
        assert orig_tx is not None, f"The original transaction {req.hash} does not exist"
        source_tx = await _trace_source(orig_tx=orig_tx)
        children = await _dfs_trace(source_tx, source_tx.hash)
        return TransactionTrace(id=source_tx.hash, transaction=source_tx, children=children)



from abc import abstractmethod
from typing import Any, Dict, Literal, Optional

import aiohttp
from aiolimiter import AsyncLimiter

__all__ = ["AsyncRequestor"]


class AsyncRequestor:
    def __init__(self, qps: float):
        self.limiter = AsyncLimiter(qps, 1)

    @abstractmethod
    def _get_request_headers(self) -> Dict[str, str]:
        raise NotImplementedError

    @abstractmethod
    async def _parse_response(self, response: aiohttp.ClientResponse):
        raise NotImplementedError

    async def _underlying_call(
        self,
        method: Literal["GET", "POST", "PUT", "DELETE"],
        url: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        payload: Optional[Dict[str, Any]] = None,
    ):
        if params:
            for k, v in params.items():
                if isinstance(v, bool):
                    params[k] = str(v).lower()
        async with self.limiter:
            async with aiohttp.ClientSession() as session:
                async with session.request(method, url=url, headers=self._get_request_headers(), params=params, json=payload) as response:
                    return await self._parse_response(response)



import asyncio
from functools import wraps

import click
import pyfiglet

from pytoncenter import get_client
from pytoncenter.utils import format_trace
from pytoncenter.v3.models import *


def asyncmd(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


def common_options(f):
    f = click.option("--network", "-n", default="testnet", help="Network", type=click.Choice(["testnet", "mainnet"]))(f)
    f = click.option("--api-key", "-k", envvar="TONCENTER_API_KEY", default=None, help="API Key")(f)
    return f


CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"],
)

WELCOME_MESSAGE = f"\033[93m{pyfiglet.figlet_format('welcome to pytoncenter')}\033[0m"
VERSION_INFO = f"\nversion  |  %(version)s \n"

# fmt: off
@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True, no_args_is_help=True,)
@click.version_option(package_name="pytoncenter", message=f"\033[93m{pyfiglet.figlet_format('welcome to pytoncenter')}\033[0m {VERSION_INFO}")
def cli():
    pass


@cli.command(help="Get transaction trace which contains the specified transaction")
@click.option("--txhash", required=True)
@common_options
@asyncmd
async def trace(txhash: str,network:Literal["mainnet","testnet"], api_key: str):
    client = get_client(version="v3", network=network) # type: ignore
    trace = await client.get_trace_alternative(GetTransactionTraceRequest(hash=txhash))
    fmt = format_trace(trace)
    click.echo(fmt)

@cli.command(help="Get account balance")
@click.option("--address", required=True, help="Address")
@common_options
@asyncmd
async def balance(address: str, network: Literal["mainnet","testnet"], api_key: str):
    client = get_client(version="v3", network=network) # type: ignore
    balance = await client.get_account(GetAccountRequest(address=address))
    click.echo(f"Address: {address}\nBalance: {balance.balance / 1e9} TON")


if __name__ == "__main__":
    cli()



from abc import ABCMeta, abstractmethod
from threading import Lock
from typing import List


class BaseKeyRotator(metaclass=ABCMeta):
    def __init__(self, keys: List[str]) -> None:
        assert isinstance(keys, list), "keys must be a list of strings"
        assert len(keys) > 0, "keys must not be empty"
        self.keys = keys

    @abstractmethod
    def get_key(self):
        raise NotImplementedError


class RoundRobinKeyRotator(BaseKeyRotator):
    def __init__(self, keys: List[str]) -> None:
        super().__init__(keys)
        self.index = 0
        self.lock = Lock()

    def get_key(self):
        with self.lock:
            key = self.keys[self.index]
            self.index = (self.index + 1) % len(self.keys)
        return key



from __future__ import annotations

import base64
import ctypes
import math
from typing import Optional, TypedDict, Union

_AddressInfo = TypedDict(
    "_AddressInfo",
    {
        "is_test_only": bool,
        "is_bounceable": bool,
        "workchain": int,
        "hash_part": bytearray,
    },
)


def string_to_bytes(string, size=1):  # ?
    if size == 1:
        buf = (ctypes.c_uint8 * len(string))()
    elif size == 2:
        buf = (ctypes.c_uint16 * len(string) * 2)()
    elif size == 4:
        buf = (ctypes.c_uint32 * len(string) * 4)()

    for i, c in enumerate(string):
        # buf[i] = ord(c)
        buf[i] = c  # ?

    return bytes(buf)


def crc16(data):
    POLY = 0x1021
    reg = 0
    message = bytes(data) + bytes(2)

    for byte in message:
        mask = 0x80
        while mask > 0:
            reg <<= 1
            if byte & mask:
                reg += 1
            mask >>= 1
            if reg > 0xFFFF:
                reg &= 0xFFFF
                reg ^= POLY

    return bytearray([math.floor(reg / 256), reg % 256])


def parse_friendly_address(addr_str: str) -> _AddressInfo:
    if len(addr_str) != 48:
        raise Exception("User-friendly address should contain strictly 48 characters")

    # avoid padding error (https://gist.github.com/perrygeo/ee7c65bb1541ff6ac770)
    data = string_to_bytes(base64.b64decode(addr_str + "=="))

    if len(data) != 36:
        raise Exception("Unknown address type: byte length is not equal to 36")

    addr = data[:34]
    crc = data[34:36]
    calced_crc = crc16(addr)
    if not (calced_crc[0] == crc[0] and calced_crc[1] == crc[1]):
        raise Exception("Wrong crc16 hashsum")

    tag = addr[0]
    is_test_only = False
    is_bounceable = False
    if tag & Address.TEST_FLAG:
        is_test_only = True
        tag ^= Address.TEST_FLAG
    if (tag != Address.BOUNCEABLE_TAG) and (tag != Address.NON_BOUNCEABLE_TAG):
        raise Exception("Unknown address tag")

    is_bounceable = tag == Address.BOUNCEABLE_TAG

    if addr[1] == 0xFF:
        workchain = -1
    else:
        workchain = addr[1]
    if workchain != 0 and workchain != -1:
        raise Exception(f"Invalid address wc {workchain}")

    hash_part = bytearray(addr[2:34])
    return {
        "is_test_only": is_test_only,
        "is_bounceable": is_bounceable,
        "workchain": workchain,
        "hash_part": hash_part,
    }


class Address:
    BOUNCEABLE_TAG = 0x11
    NON_BOUNCEABLE_TAG = 0x51
    TEST_FLAG = 0x80

    def __init__(self, any_form: Union[str, Address]):
        if any_form is None:
            raise Exception("Invalid address")

        if isinstance(any_form, Address):
            self._wc = any_form._wc
            self._hash_part = any_form._hash_part
            self._is_test_only = any_form._is_test_only
            self._is_user_friendly = any_form._is_user_friendly
            self._is_bounceable = any_form._is_bounceable
            self._is_url_safe = any_form._is_url_safe
            return

        # base64 with digits, upper and lowercase Latin letters, '/' and '+'
        # base64url with '_' and '-' instead of '/' and '+')
        if any_form.find("-") > 0 or any_form.find("_") > 0:
            any_form = any_form.replace("-", "+").replace("_", "/")
            self._is_url_safe = True
        else:
            self._is_url_safe = False

        # Check this address is raw address or not
        try:
            colon_index = any_form.index(":")
        except ValueError:
            colon_index = -1  # raw address

        if colon_index > -1:
            # Example of Raw Address, -1:fcb91a3a3816d0f7b8c2c76108b8a9bc5a6b7a55bd79f8ab101c52db29232260
            arr = any_form.split(":")
            if len(arr) != 2:
                raise Exception(f"Invalid address {any_form}")

            wc = int(arr[0])
            if wc != 0 and wc != -1:
                raise Exception(f"Invalid address wc {wc}")

            address_hex = arr[1]
            if len(address_hex) != 64:
                raise Exception(f"Invalid address hex {any_form}")

            self._is_user_friendly = False
            self._wc = wc
            self._hash_part = bytearray.fromhex(address_hex)
            self._is_test_only = False
            self._is_bounceable = False
        else:
            self._is_user_friendly = True
            parse_result = parse_friendly_address(any_form)
            self._wc = parse_result["workchain"]
            self._hash_part = parse_result["hash_part"]
            self._is_test_only = parse_result["is_test_only"]
            self._is_bounceable = parse_result["is_bounceable"]

    @property
    def wc(self) -> int:
        """
        Alias for workchain, compatible with tonsdk
        """
        return self._wc

    @property
    def workchain(self) -> int:
        return self._wc

    @property
    def hash_part(self) -> bytearray:
        return self._hash_part

    @property
    def is_test_only(self) -> bool:
        return self._is_test_only

    @property
    def is_user_friendly(self) -> bool:
        return self._is_user_friendly

    @property
    def is_bounceable(self) -> bool:
        return self._is_bounceable

    @property
    def is_url_safe(self) -> bool:
        return self._is_url_safe

    def to_string(
        self,
        is_user_friendly: Optional[bool] = None,
        is_url_safe: Optional[bool] = None,
        is_bounceable: Optional[bool] = None,
        is_test_only: Optional[bool] = None,
    ) -> str:
        if is_user_friendly is None:
            is_user_friendly = self._is_user_friendly
        if is_url_safe is None:
            is_url_safe = self._is_url_safe
        if is_bounceable is None:
            is_bounceable = self._is_bounceable
        if is_test_only is None:
            is_test_only = self._is_test_only

        if not is_user_friendly:
            return f"{self._wc}:{self._hash_part.hex()}"
        else:
            tag = Address.BOUNCEABLE_TAG if is_bounceable else Address.NON_BOUNCEABLE_TAG

            if is_test_only:
                tag |= Address.TEST_FLAG

            addr = (ctypes.c_int8 * 34)()
            addr[0] = tag
            addr[1] = self._wc
            addr[2:] = self._hash_part
            address_with_checksum = (ctypes.c_uint8 * 36)()
            address_with_checksum[:34] = addr
            address_with_checksum[34:] = crc16(addr)

            address_base_64 = base64.b64encode(address_with_checksum).decode("utf-8")
            if is_url_safe:
                address_base_64 = address_base_64.replace("+", "-").replace("/", "_")

            return str(address_base_64)

    def __hash__(self) -> int:
        return hash(self.to_string(True, True, True, is_test_only=False))

    def __key__(self) -> str:
        return self.to_string(True, True, True, is_test_only=False)

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, Address):
            return self.to_string(True, True, True, is_test_only=False) == __value.to_string(True, True, True, is_test_only=False)
        if isinstance(__value, str):
            return self.to_string(True, True, True, is_test_only=False) == Address(__value).to_string(True, True, True, is_test_only=False)
        return False

    def __repr__(self) -> str:
        return self.to_string(True, True, True, is_test_only=False)
